@@ -6,7 +6,7 @@
 from pwd import getpwnam, getpwuid
 from piraha import parse_peg_src, Matcher, Group, set_trace
 from subprocess import Popen, PIPE, STDOUT
-from .pipe_threads import PipeThread, get_lastpid, get_running
+from .pipe_threads import PipeThread, get_lastpid, get_running, pwait
 import os
 import sys
 import re
@@ -17,6 +17,29 @@ from shutil import which
 from datetime import datetime
 from .tmpfile import tmpfile
 from .version import __version__
+from time import time
+import json
+
+def prepJson(arg):
+    if arg is None:
+        return []
+    t = type(arg)
+    if t == dict:
+        narg = {}
+        for k in arg:
+            narg[k] = prepJson(arg[k])
+    elif t == list:
+        narg = []
+        for k in range(len(arg)):
+            narg += [prepJson(arg[k])]
+    elif t == set:
+        narg = prepJson(list(arg))
+    elif t in [int,float,bool]:
+        narg = arg
+    else:
+        narg = str(arg)
+    return narg
+
 
 # The way exit works is to raise SystemExit,
 # which may be caught. When we want to exit our
@@ -110,19 +133,21 @@ grammar = r"""
 skipper=\b([ \t]|\\\n|\#.*)*
 s=\b([ \t\n]|\\\n|\#.*)*
 raw_word=(\\.|[^\\"'\t \n\$\#&\|;{}`()<>?*,])+
+raw_word2=(\\.|[^()[\]{}]|\[{raw_word2}\]|\({raw_word2}\)|\{{raw_word2}\})*
 dchar=[^"$\\]
 dlit=\\.
 dquote="({dlit}|{dchar}|{var}|{math}|{subproc}|{bquote})*"
 bquote=`(\\.|[^`$]|{var}|{math}|{subproc})*`
 squote='(\\.|[^'])*'
 unset=:-
+unsetp=:\+
 unset_raw=-
 rm_front=\#\#?
 rm_back=%%?
 def=:-
 wchar=[@?$!-]
 w=[a-zA-Z0-9_]+
-var=\$({wchar}|{w}|\{(({w}|{wchar})({unset}{words2}|{unset_raw}{raw_word}|{rm_front}{word2}|{rm_back}{word2}|{def}{word2}?|))\})
+var=\$({wchar}|{w}|\{(({w}|{wchar})({unset}{words2}|{unsetp}{raw_word2}|{unset_raw}{raw_word}|{rm_front}{word2}|{rm_back}{word2}|{def}{word2}?|))\})
 func=function {ident} \( \) \{( {cmd})* \}({redir}|[ \t])*\n
 
 worditem=({glob}|{redir}|{var}|{math}|{subproc}|{raw_word}|{squote}|{dquote}|{bquote})
@@ -139,7 +164,7 @@ case=case {word} in({-s}{casepattern}|)
 case2=;;{-s}({esac}|{casepattern}|)
 
 fd_from=[0-9]+
-fd_to=&[0-9]+
+fd_to=&[0-9]*
 ltgt=(<<<|<<|<|>>|>)
 redir=({fd_from}|){ltgt}( {fd_to}| {word})
 ident=[a-zA-Z0-9][a-zA-Z0-9_]*
@@ -397,6 +422,7 @@ class shell:
             "PWD":os.path.realpath(os.getcwd()),
             "*":" ".join(self.args[1:]),
             "SHELL":os.path.realpath(shell_name),
+            "PYTHON":sys.executable,
             "EVSHELL":__version__}
         pwdata = getpwuid(os.getuid())
         self.vars["USER"] = pwdata.pw_name
@@ -430,7 +456,11 @@ class shell:
         self.last_pipe = None
         self.recursion = 0
         self.max_recursion_depth = 20
-        self.log_fd = open(os.path.join(self.vars["HOME"],"evshell-log.txt"),"a")
+        log_file_dir = os.path.join(self.vars["HOME"],".evshell-logs")
+        os.makedirs(log_file_dir, exist_ok = True)
+        log_file = os.path.join(log_file_dir, f"log-{os.getpid()}.jtxt")
+        self.log_fd = open(log_file,"w")
+        self.log(msg="starting shell")
 
     def err(self, e):
         print(colored(str(e),"red"),file=self.stderr)
@@ -438,17 +468,21 @@ class shell:
     def open_file(self, fname, rwa, line):
         sout = None
         try:
-            return open(fname, rwa)
+            sout = open(fname, rwa)
+            self.log(open=fname,rwa=rwa)
         except PermissionError as pe:
+            self.log(exc=pe,open=fname,rwa=rwa)
             print(self.scriptname,":",
                 " line ", line, ": ", fname,": ",
                 pe.strerror.strip(),
                 sep="", file=self.stderr)
         except OSError as e:
+            self.log(exc=pe,open=fname,rwa=rwa)
             self.err(e)
         if sout is None:
             return open("/dev/null", rwa)
-        self.stdrr.flush()
+        if self.stderr is not None:
+            self.stderr.flush()
         return sout
     
     def env_is_bound(self):
@@ -470,30 +504,36 @@ class shell:
     def log_flush(self):
         self.log_fd.flush()
 
-    def log_exc(self):
+    def log_exc(self,e):
         print_exc()
-        print("=" * 10,datetime.now(),"=" * 10,file=self.log_fd)
         self.log_flush()
-        print_exc(file=self.log_fd)
+        self.log(exc=e)
         self.log_flush()
 
-    def log(self,*args,**kwargs):
-        print("=" * 10,datetime.now(),"=" * 10,file=self.log_fd)
+    def log(self,**kwargs):
         self.log_flush()
-        print(*args,**kwargs,file=self.log_fd)
+        args = prepJson(kwargs)
+        if "time" not in args:
+            args["time"] = time()
+        args["script"] = self.scriptname
+        print(json.dumps(args),file=self.log_fd)
         self.log_flush()
 
     def unset_var(self,vname):
         val = self.allow_set_var(vname, None)
         if val == None:
+            unset = False
             try:
                 del self.vars[vname]
+                unset = True
             except KeyError as ke:
                 pass
             try:
                 del self.exports[vname]
+                unset = True
             except KeyError as ke:
                 pass
+            self.log(unset=vname)
 
     def get_var(self,vname):
         if vname in self.exports:
@@ -502,6 +542,7 @@ class shell:
             return self.vars.get(vname,"")
 
     def set_var(self,vname,value):
+        self.log(setvar=vname,value=value)
         assert vname != "2"
         value = self.allow_set_var(vname, value)
         self.vars[vname] = value
@@ -842,7 +883,7 @@ class shell:
                 self.stdout.write(out)
             rc=os.waitpid(pid,0)
             self.vars["?"] = str(rc[1])
-            self.log("end subshell:",self.vars["?"])
+            self.log(msg="end subshell",rc=self.vars["?"])
             if self.vars["?"] != "0" and self.flags.get("e",False):
                 shell_exit(int(self.vars["?"]))
             return []
@@ -937,7 +978,7 @@ class shell:
                             varname = g.group(1)
                             value = g.group(2)
                             self.set_var(varname,value)
-                            self.exports[varname] = value
+                            self.exports[varname] = self.vars[varname]
                         elif a in self.vars:
                             self.exports[a] = self.vars[a]
                     return
@@ -1026,15 +1067,9 @@ class shell:
                 return []
             if args[0] == "wait":
                 result = None
-                try:
-                    pid, status = os.wait()
-                    p = get_running(pid)
-                    result = p.communicate()
-                except Exception as ee:
-                    p = get_running(None)
-                    if p is not None:
-                        result = p.communicate()
-                self.log("end wait:",self.vars["?"])
+                p = pwait(None)
+                print("pid:",p.getpid(),"cmd:",p.args[0])
+                self.log(msg="end wait",pid=p.getpid(),rc=p.returncode)
                 return []
             if args[0] == "cd":
                 if len(args) == 1:
@@ -1043,7 +1078,7 @@ class shell:
                     cd_dir = args[1]
                 cd_dir = self.allow_cd(cd_dir)
                 os.chdir(cd_dir)
-                self.log("chdir:",cd_dir)
+                self.log(chdir=cd_dir)
                 self.vars["PWD"] = os.getcwd()
                 return
 
@@ -1139,24 +1174,30 @@ class shell:
                         if first_line.startswith("#!"):
                             args = re.split(r'\s+',first_line[2:].strip()) + args
                 if not os.path.exists(args[0]):
-                    fno = gr.linenum()
+                    if gr is None:
+                        fno = 0
+                    else:
+                        fno = gr.linenum()
                     print(f"{self.scriptname}: line {fno}: {args[0]}: command not found",file=self.stderr)
+                    self.log(args=args,msg="command not found",line=fno)
                     self.vars["?"] = 1
                     if self.flags.get("e",False):
                         shell_exit(1)
                     return ""
                 args = self.allow_cmd(args)
-                self.log("args:",args)
                 env = {}
                 if self.flags.get("x",False):
                     self.stderr.write("+ "+" ".join(args)+"\n")
                 for e in self.exports:
                     env[e] = self.exports[e]
                 try:
+                    tstart = time()
                     p = PipeThread(args, stdin=sin, stdout=sout, stderr=serr, universal_newlines=True, env=env)
+                    self.log(msg="start",pid=p.getpid(), args=args, time=tstart)
                 except OSError as e:
                     args = ["/bin/sh"]+args
                     p = PipeThread(args, stdin=sin, stdout=sout, stderr=serr, universal_newlines=True, env=env)
+                    self.log(msg="start",pid=p.getpid(), args=args)
                 if self.curr_ending == "&":
                     p.background()
                     p.start()
@@ -1168,7 +1209,7 @@ class shell:
                     p.start()
                     p.communicate()
                     self.vars["?"] = str(p.returncode)
-                    self.log("end cmd:",self.vars["?"])
+                    self.log(msg="end", rc=self.vars["?"], pid=p.getpid())
                     if self.vars["?"] != "0" and self.flags.get("e",False):
                         shell_exit(int(self.vars["?"]))
                     return None
@@ -1215,7 +1256,11 @@ class shell:
 
     def run_text_(self,txt):
         #here(colored("="*50,"yellow"))
-        self.log("txt:",txt)
+        nchars=50
+        if len(txt) < nchars:
+            self.log(txt=txt,full=True)
+        else:
+            self.log(txt=txt[0:nchars]+"...",full=False)
         txt = self.txt + txt
         if txt.endswith("\\\n"):
             self.txt = txt
@@ -1255,14 +1300,14 @@ class shell:
             #print()
             #m.showError()
             #print("continue...")
-            self.log("CONTINUE")
+            self.log(msg="CONTINUE")
             return "CONTNUE"
         else:
             self.txt = ''
             m.showError()
             here("done")
-            self.log("SYNTAX")
-            m.showError(self.log_fd)
+            self.log(msg="SYNTAX")
+            #m.showError(self.log_fd)
             return "SYNTAX"
 
 def interactive(shell):
@@ -1290,16 +1335,19 @@ def interactive(shell):
 def run_interactive(s):
     try:
         rc = interactive(s)
-        s.log("rc2:",rc)
+        s.log(msg="session ended normally",rc=rc)
     except ShellAccess as sa:
-        self.err(sa,self.stderr)
+        s.err(sa) #,s.stderr)
         rc = -1
+        s.log(msg="session ended with access error",rc=rc)
     except ShellExit as se:
         rc = se.rc
+        s.log(msg="session ended normally",rc=rc)
         exit(rc)
     except Exception as ee:
         rc = 1
-        s.log_exc()
+        s.log(msg="session ended with exception",rc=rc,exc=ee)
+        s.log_exc(ee)
     exit(rc)
 
 def run_shell(s):
@@ -1308,14 +1356,14 @@ def run_shell(s):
     if ssh_cmd is not None:
         try:
             rc = s.run_text(ssh_cmd)
-            s.log("rc1:",rc)
+            s.log(rc=rc)
         except Exception as ee:
-            s.log_exc()
-    elif s.shell_name != s.args[0]:
+            s.log_exc(ee)
+    elif os.path.realpath(s.shell_name) != os.path.realpath(s.args[0]):
         s.scriptname  = s.args[0]
         with s.open_file(s.args[0],"r",1) as fd:
             rc = s.run_text(fd.read())
-            s.log("rc5",rc)
+            s.log(rc=rc)
     else:
         found = False
         for n in range(1,len(args)):
@@ -1329,17 +1377,17 @@ def run_shell(s):
                     try:
                         found = True
                         rc = s.run_text(fd.read())
-                        s.log("rc3:",rc)
+                        s.log(rc=rc)
                         assert rc == "EVAL", f"rc={rc}"
                     except ShellAccess as sa:
                         rc = -1
-                        self.err(sa,self.stderr)
+                        s.err(sa)
                         exit(rc)
                     except ShellExit as se:
                         rc = se.rc
                         exit(rc)
                     except Exception as ee:
-                        s.log_exc()
+                        s.log_exc(ee)
         if not found:
             run_interactive(s)
 
